﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public enum CompDevHazard {
    None,
    Stop,
    Boost,
    Leech,
    Pulse,
    Teleport,
}

[RequireComponent(typeof(SpriteRenderer))]
public class Hazard : MonoBehaviour {
    public const int POINT_VALUE = 500;

    public const int LEECH_SUBTRACT_VALUE = 250;

    private const string PLAYER_TAG = "Player";

    // Hazard will not trigger when its owner collides with it
    public GameObject Owner;
    public CompDevHazard HazardType;
	public GameObject EffectPrefab;

	// Degrees per second
	private const float ROTATION_SPEED = 150f;
    // Some hazards may rotate
    protected float Rotation;

    // Use this for initialization
    void Awake() {
        Rotation = 0;
    }

    void Update() {
        if (HazardType == CompDevHazard.Boost) {
            Rotation = (Rotation + ROTATION_SPEED * Time.deltaTime) % 360;
            transform.rotation = Quaternion.Euler(new Vector3(0f, 0f, Rotation));
        }
    }

    public void Init(GameObject owner, CompDevHazard hazardType) {
        Owner = owner;
        GetComponent<SpriteRenderer>().color = owner.transform.Find("Sprite").GetComponent<SpriteRenderer>().color;
        HazardType = hazardType;
    }

    void OnTriggerEnter2D(Collider2D other) {
        if (other.tag == PLAYER_TAG && other.gameObject != Owner) {
            Trigger(other.gameObject);
            Destroy(gameObject);
        }
    }

    void Trigger(GameObject other) {
        PlayerMovement ownerPlayer = Owner.GetComponent<PlayerMovement>();
        float dashForce = other.GetComponent<PlayerMovement>().DashForce;

        Rigidbody2D otherRB = other.GetComponent<Rigidbody2D>();
        switch (HazardType) {
        case CompDevHazard.Stop:
            otherRB.velocity = Vector2.zero;
            break;
        case CompDevHazard.Boost:
            Vector2 boost = Quaternion.AngleAxis(Rotation, Vector3.forward) * Vector3.right;
            boost *= dashForce;
            otherRB.AddForce(boost, ForceMode2D.Impulse);
            break;
        case CompDevHazard.Leech:
            if (ownerPlayer.GamePadIndex == 0) {
                GameController.Player2Score -= Hazard.LEECH_SUBTRACT_VALUE;
            }
            else if (ownerPlayer.GamePadIndex == 1) {
                GameController.Player1Score -= Hazard.LEECH_SUBTRACT_VALUE;
            }
            break;
        case CompDevHazard.Pulse:
            Vector2 otherPosition = other.transform.Find("Sprite")
                .GetComponent<SpriteRenderer>()
                .transform.position;
            Vector2 myPosition = new Vector2(transform.position.x, transform.position.y);
            Vector2 pulse = (otherPosition - myPosition).normalized * dashForce;
            otherRB.AddForce(pulse, ForceMode2D.Impulse);
            break;
        case CompDevHazard.Teleport:
            Vector2 teleportPosition = GameController.GetValidSpawnPoint();
            other.transform.position = teleportPosition;
            if (EffectPrefab != null)
            {
				Instantiate(EffectPrefab, new Vector3(teleportPosition.x, teleportPosition.y, 0f), Quaternion.identity);
			}
            break;
        }


        if (ownerPlayer.GamePadIndex == 0) {
            GameController.Player1Score += Hazard.POINT_VALUE;
        }
        else if (ownerPlayer.GamePadIndex == 1) {
            GameController.Player2Score += Hazard.POINT_VALUE;
        }
    }
}
