﻿using System.Collections;
using UnityEngine;

public class PlayerMovement : MonoBehaviour
{
    // Whether the pedal was pressed in the last fixed update
    protected bool PedalWasPressed;
    // Whether the pedal is pressed in the current fixed update
    protected bool PedalNowPressed;
    // Whether dash was pressed in the last fixed update
    protected bool DashWasPressed;
    // Whether the dash is pressed in the current fixed update
    protected bool DashNowPressed;

    public float DashForce = 400.0f;
    // How long after dashing before movement input is registered again
    public float DashCooldown = 1.0f;

    public float PedalForce = 100.0f;
    public float MaxSpeed = 500.0f;
	public SpriteRenderer PlayerSprite = null;
	public int GamePadIndex = 0;

    public AudioSource DashSound;
    public AudioSource HazardSound;
    public AudioSource PlayerSound;
    public AudioSource WallSound;

	private Rigidbody2D rb;
	//private Camera MainCamera = null;
	//private Bounds CameraBounds = new Bounds();
    private float CurDashCooldown;

    private bool Frozen;
    private float FrozenCountdown;

	void Awake()
    {
        PedalWasPressed = false;
		//MainCamera = Camera.main;
        // float verticalHeightSeen = MainCamera.orthographicSize * 2f;
		// CameraBounds.size = new Vector3((verticalHeightSeen * MainCamera.aspect), verticalHeightSeen, 0);
        rb = GetComponent<Rigidbody2D>();

        CurDashCooldown = 0.0f;

        FrozenCountdown = 0f;
    }

    void Start()
    {
		
	}

    public void Freeze(float duration)
    {
        FrozenCountdown = duration;
        rb.velocity = Vector2.zero;
    }

    void FixedUpdate()
    {
        if (GameController.GameStart == false)
        {
			return;
		}

        if (FrozenCountdown > 0f) {
            FrozenCountdown -= Time.deltaTime;
            return;
        }

        // Update the input for the current frame
        DashNowPressed = InputManager.Instance
            .GetCompDevInputButton(CompDevInput.Dash, GamePadIndex);
		PedalNowPressed = InputManager.Instance
			.GetCompDevInputButton(CompDevInput.Move, GamePadIndex);

        Vector2 force = Vector2.zero;

        // Only allow input to influence movement if we're not
        // in the middle of a dash cooldown
        if (CurDashCooldown > 0.0f) {
            CurDashCooldown -= Time.deltaTime;
        }
        else {
            // Check for a dash first; that takes priority over pedaling
            Vector2 dashForce = GetDashForce();

            if (dashForce.magnitude == 0) 
            {
                force = GetPedalForce();
            }
            else
            {
                rb.AddForce(dashForce, ForceMode2D.Impulse);
                if (DashSound.isPlaying == false)
				{
					DashSound.Play();
				}
			}
        }

		rb.AddForce(force);

        if (rb.velocity.magnitude > MaxSpeed)
        {
			rb.velocity.Normalize();
			rb.velocity *= MaxSpeed;
		}

        // Update the current input for the next update
        if (DashNowPressed) {
            DashWasPressed = true;
        }
        else {
            DashWasPressed = false;
        }
        if (PedalNowPressed) {
            PedalWasPressed = true;
        }
        else {
            PedalWasPressed = false;
        }
    }

    // Dashing can only apply in the current direction of movement
    // and starts an input cooldown.
    Vector2 GetDashForce() {
		float dashMoveComponent = 0;
        if (DashNowPressed && !DashWasPressed) {
            dashMoveComponent = DashForce;
            // Start the cooldown, since we just recorded
            // dash input
            CurDashCooldown = DashCooldown;
        }

        Vector2 dashForce = rb.velocity.normalized;
        dashForce *= dashMoveComponent;

        return dashForce;
    }

    // Pedaling can apply in any direction and isn't subject to a cooldown.
    Vector2 GetPedalForce() {
        float inputXComponent = InputManager.Instance
            .GetCompDevInputAxis(CompDevInput.DirectionHorizontal, GamePadIndex);

        float inputYComponent = InputManager.Instance
            .GetCompDevInputAxis(CompDevInput.DirectionVertical, GamePadIndex);

        Vector2 force = new Vector2(inputXComponent, inputYComponent);

		float pedalMoveComponent = 0;
        if (PedalNowPressed && !PedalWasPressed) {
            pedalMoveComponent = PedalForce;
        }

		force.Normalize();
		force *= pedalMoveComponent;

        return force;

    }

    void LateUpdate()
	{
        Vector3 position = transform.position;

		Bounds gridBounds = GameController.GridBounds;

		if (position.x > gridBounds.max.x)
        {
			position.x = gridBounds.min.x;
		}
        else if (position.y > gridBounds.max.y)
        {
            position.y = gridBounds.min.y;
        }

        if (position.x < gridBounds.min.x)
		{
            position.x = gridBounds.max.x;
		}
		else if (position.y < gridBounds.min.y)
		{
            position.y = gridBounds.max.y;
		}

		transform.position = position;

        if (PlayerSprite != null)
		{
			Vector2 normalVelocity = rb.velocity.normalized;
			float angle = Mathf.Acos(normalVelocity.x) * Mathf.Rad2Deg;
            if (normalVelocity.y < 0)
            {
				angle *= -1;
			}
			PlayerSprite.transform.rotation = Quaternion.Euler(new Vector3(0f, 0f, angle));
		}
	}

	void OnCollisionEnter2D(Collision2D other)
	{
        PlayerMovement otherPlayerMovement = other.gameObject.GetComponent<PlayerMovement>();

		if (otherPlayerMovement != null &&
            PlayerSound.isPlaying == false)
		{
			PlayerSound.Play();
		}

        BoxCollider2D otherBoxCollider = other.gameObject.GetComponent<BoxCollider2D>();

		if (otherBoxCollider != null &&
			WallSound.isPlaying == false)
		{
			WallSound.Play();
		}
	}

	void OnTriggerEnter2D(Collider2D other)
	{
        Hazard otherHazard = other.gameObject.GetComponent<Hazard>();

		if (otherHazard != null &&
            otherHazard.Owner != gameObject &&
			HazardSound.isPlaying == false)
		{
			HazardSound.Play();
		}
	}
}
